
/* Sx cpu micro architecture simulator

   5 December 2001
   modify from s2m sim
   Computer Architecture 1997, 1998
   microprogram version 		13 Feb 2005

   sx micro						10 July 2006
   sx0  minimum version			11 July 2006

   integrate to noss			15 Aug 2006

   P. Chongstitvatana
   Department of Computer Engineering
   Chulalongkorn University
*/

#include "sx.h"

#define TIMEOUT  1000	// time quanta

#define time_out 10		// process status
#define stopped  11
#define blocked  12

extern int CP, DP;		// load.c
extern int ustart;

int status;				// process status
int userp;				// pdes of current user process
int intflag;			// interrupt flag

void initchip(void){
	clock = 0;
	ninst = 0;
	runflag = 1;
	TS = 0;
	NX = 0;
	FF = 0;
	FP = SSBASE;
	SP = SSBASE;
	PC = 1;
	Z = 0;
	S = 0;
	mpc = 0;
}

void decode(int code, int *op, int *arg){
	*op = code & 255;
	*arg = code >> 8;
}

PRIVATE void popts(void){
	TS = M[SP];
	SP--;
}

// static alloc memory of size a
PRIVATE int xalloc(int a){
	int b;
	b = DP;
	if( DP + a >= MAXDS )
		error("xalloc: out of memory");
	DP += a;
	return b;
}

PRIVATE void report(void){
  float cpi;
  if( ninst == 0 )
	printf("0 instruction\n");
  else {
	cpi = ((float) clock)/ninst;
	printf("%d instructions, %d cycles, CPI %4.2f\n",
		ninst,clock,cpi);
  }
}

// save c-state to process descriptor a
// not really a user function
// call by boot() after time-out
void saveCstate(int a){
	M[a + 4] = FP;
	M[a + 5] = SP;
	M[a + 6] = PC;
	M[a + 7] = TS;
}
// restore c-state from a process desc.
// call by "runnable" in user space
PRIVATE void restoreCstate(int a){
	userp = a;
	FP = M[a + 4];
	SP = M[a + 5];
	PC = M[a + 6];	// cause a jump
	TS = M[a + 7];
	ustart = clock;	// start of user time
}

PRIVATE void trace(void){
	int op, arg;
	decode(M[PC],&op,&arg);
//	if(op == icCall)
//		printf("call %s\n",searchref(tyFUN,arg));
	prCode(stdout,M[PC],PC);
	printf(" ");
	if(op == icCall) printf("\n");
}

// run until timeout/blocked/stopped
void eval(int ref){
	clock = 0;
	runflag = 1;
	PC = ref;
	while(runflag ) {
		if( intflag )
			if( clock > TIMEOUT ){
			status = time_out;
			break;
			}
//		trace();
		run3();		// execute one instruction
		ninst++;
	}
	saveCstate(userp);
}

void trap(int op, int arg){		// special functions
	int n, a;
	switch(op) {
	case icEnd :
//		report();
		status = stopped;
		runflag = 0;
		break;
	case icArray:
		TS = xalloc(TS);
		break;
	case icSys:
		switch(arg){
		case 1: 				// print ts
			printf("%d",TS);
			popts();
			break;
		case 2 : 				// printc ts
			printf("%c",TS);
			popts();
			break;
		case 13:
			runflag = 0;
			break;
		case 20: intflag = 1; break;
		case 21: intflag = 0; break;
		case 22:
			status = blocked;
			runflag = 0;
			break;
/*
		case 23:
			a = TS;				// pdes
			popts();
			saveCstate(a);
			break;
*/
		case 24:			// restore C-state
			a = TS;
			popts();
			restoreCstate(a);
			intflag = 1;
			break;
		default:
			error("unknown syscall");
		}
		break;
	default:
		error("unknown trap");
	}
}

/*
int main(int argc, char *argv[]){
  if( argc < 2 ) {
	printf("usage : sx objfile\n");
	exit(0);
  }
  loadobj(argv[1]);
  initchip();
  run();
  return 0;
}
*/
